package aerogen;

import java.util.ArrayList;
import java.util.Random;

public class CaveSystem {
	
	ArrayList<CaveNode> currentEnds;
	
	ArrayList<CaveNode>[][][] nodeGrid;
	int gridLength, gridHeight, gridBredth;
	
	CaveDescription description;
	CaveNode startingNode;
	double length, height, bredth;
	double splits = 1;
	int segments = 0;
	Random random;
	int stayBelow;
	Island island;
	
	@SuppressWarnings("unchecked")
	public CaveSystem(Island island, Random random, CaveNode start, int maxCaveSpace) {
		this.startingNode = start;
		this.currentEnds = new ArrayList<CaveNode>();
		
		this.island = island;		
		this.length = island.length;
		this.bredth = island.bredth;
		this.height = island.height;
		this.description = island.biome.caveDescription;
		this.stayBelow = (int) (island.centerY+island.biome.roughness/2);
		this.random = random;
		
		gridLength = ceil((double)length/8.0);
		gridHeight = ceil((double)height/8.0);
		gridBredth = ceil((double)bredth/8.0);
		this.nodeGrid = new ArrayList[gridLength][gridHeight][gridBredth];
		for(int i = 0; i < gridLength; i++) {
			for(int j = 0; j < gridHeight; j++) {
				for(int k = 0; k < gridBredth; k++) {
					nodeGrid[i][j][k] = new ArrayList<CaveNode>();
				}
			}
		}
		this.currentEnds.add(start);
		addToGrid(start);
		
		int minSegments = (int) ((double)maxCaveSpace / (Math.PI*description.startRadius*description.startRadius)*2.0);
		while(this.segments < minSegments && currentEnds.size() > 0) {
			expand();
		}
		
	}
	
	private int ceil(double d) {
		int i = 0;
		if(d < 0) i = (int) d - 1;
		if(d > 0) i = (int) d;
		if(i == d) return i;
		return i + 1;
	}
	
	private int floor(double d) {
		return ceil(d)-1;
	}

	public void expand() {
		int selectedIndex = random.nextInt(currentEnds.size());
		CaveNode selectedEnd = currentEnds.get(selectedIndex);
		currentEnds.remove(selectedIndex);
		
		int number = 1;
		while(number < description.maxSplit) {
			if(random.nextDouble() < description.splitChance/splits) {
				number++;
				splits++;
			} else {
				break;
			}
		}
		double swayDX = description.angleDX*number*number;
		double swayDY = description.angleDY*number*number;
		
		ArrayList<CaveNode> toAdd = new ArrayList<CaveNode>(); 
		for(int i = 0; i < number; i++) {
			
			double x = selectedEnd.x + Math.cos(selectedEnd.angleX)*Math.cos(selectedEnd.angleY)*2.0;
			double z = selectedEnd.z + Math.sin(selectedEnd.angleX)*Math.cos(selectedEnd.angleY)*2.0;
			double y = selectedEnd.y + Math.sin(selectedEnd.angleY)*2.0;
			
			if(x >= 0 && y >= 0 && z >= 0 && x < this.length && y < this.height && y < stayBelow && z < this.bredth && island.getBlockSafe((int)x, (int)y, (int)z) != 0) {
				
				double radius = selectedEnd.radius + (random.nextDouble()*2.0-1.0)*description.maxDeltaRadius;
				if(radius < description.minRadius) radius = description.minRadius;
				if(radius > description.maxRadius) radius = description.maxRadius;

				double angleX;
				double angleY;
				if(number > 1) {
					double a = random.nextDouble()*Math.PI*2.0;
					angleX = selectedEnd.angleX + Math.cos(a)*swayDX;
					angleY = selectedEnd.angleY + Math.sin(a)*swayDY;
				} else {
					angleX = selectedEnd.angleX + (random.nextDouble()*2.0-1.0)*swayDX;
					angleY = selectedEnd.angleY + (random.nextDouble()*2.0-1.0)*swayDY;
				}
				
				if(angleY < -description.maxAngleY) angleY = -description.maxAngleY;
				if(angleY > description.maxAngleY) angleY = description.maxAngleY;
				
				CaveNode node = new CaveNode(x, y, z, angleX, angleY, radius);
				toAdd.add(node);
				currentEnds.add(node);
				addToGrid(node);
				segments += 1;
			}
		}
		
		CaveNode[] nodes = new CaveNode[toAdd.size()];
		for(int i = 0; i < nodes.length; i++) {
			nodes[i] = toAdd.get(i);
		}
		
	}
	
	private void addToGrid(CaveNode node) {
		int x = floor(node.x / 8.0);
		int y = floor(node.y / 8.0);
		int z = floor(node.z / 8.0);
		nodeGrid[x][y][z].add(node);
	}

	public boolean contains(double x, double y, double z) {
		int gx = floor(x / 8.0);
		int gy = floor(y / 8.0);
		int gz = floor(z / 8.0);
		for(int i = -1; i <= 1; i++) {
			for(int j = -1; j <= 1; j++) {
				for(int k = -1; k <= 1; k++) {
					int ix = gx+i;
					int iy = gy+j;
					int iz = gz+k;
					if(ix >= 0 && iy >= 0 && iz >= 0 && ix < gridLength && iy < gridHeight && iz < gridBredth) {
						for(CaveNode n : nodeGrid[ix][iy][iz]) {
							if(distance2(x, y, z, n.x, n.y, n.z) < n.radius2) return true;
						}
					}
				}
			}
		}
		return false;
	}

	private double distance2(double x1, double y1, double z1, double x2, double y2, double z2) {
		double xdist = x2-x1;
		double ydist = y2-y1;
		double zdist = z2-z1;
		return xdist*xdist+ydist*ydist+zdist*zdist;
	}
	
}